@@ -43,7 +43,7 @@ from .check_functions import (
     check_wide_relationship,
     check_parameter_definition,
     check_parameter_value,
-    check_wide_parameter_value_list,
+    check_parameter_value_list,
 )
 from .import_functions import (
     import_alternatives,
@@ -110,7 +110,6 @@ from .parameter_value import (
     TimeSeries,
     TimeSeriesFixedResolution,
     TimeSeriesVariableResolution,
-    ParameterValueEncoder,
 )
 from .filters.alternative_filter import apply_alternative_filter_to_parameter_value_sq
 from .filters.scenario_filter import apply_scenario_filter_to_parameter_value_sq
@@ -128,4 +127,4 @@ from .filters.tools import (
 )
 
 name = "spinedb_api"
-__version__ = '0.13.0'
+__version__ = "0.17.0"
