@@ -131,7 +131,7 @@ def check_object(item, current_items, object_class_ids, object_entity_type):
         name = item["name"]
     except KeyError:
         raise SpineIntegrityError(
-            "Python KeyError: There is no dictionary key for the object name. Probably a bug, " "please report."
+            "Python KeyError: There is no dictionary key for the object name. Probably a bug, please report."
         )
     if not name:
         raise SpineIntegrityError("Object name is an empty string and therefore not valid")
@@ -179,7 +179,7 @@ def check_wide_relationship_class(wide_item, current_items, object_class_ids, re
         )
     if not given_object_class_id_list:
         raise SpineIntegrityError(f"At least one object class is needed for the relationship class '{name}'.")
-    if not all([id_ in object_class_ids for id_ in given_object_class_id_list]):
+    if not all(id_ in object_class_ids for id_ in given_object_class_id_list):
         raise SpineIntegrityError(
             f"At least one of the object class ids of the relationship class '{name}' is not in the database."
         )
@@ -240,7 +240,7 @@ def check_wide_relationship(
     except KeyError:
         raise SpineIntegrityError(f"There is no object class id list for relationship '{name}'")
     try:
-        object_id_list = wide_item["object_id_list"]
+        object_id_list = tuple(wide_item["object_id_list"])
     except KeyError:
         raise SpineIntegrityError(f"There is no object id list for relationship '{name}'")
     try:
@@ -253,15 +253,14 @@ def check_wide_relationship(
         raise SpineIntegrityError(
             f"Incorrect objects '{object_name_list}' for relationship class '{relationship_class_name}'."
         )
-    join_object_id_list = ",".join([str(x) for x in object_id_list])
-    if (class_id, join_object_id_list) in current_items_by_obj_lst:
+    if (class_id, object_id_list) in current_items_by_obj_lst:
         object_name_list = [objects[id]["name"] for id in object_id_list]
         relationship_class_name = relationship_classes[class_id]["name"]
         raise SpineIntegrityError(
             "There's already a relationship between objects {} in class {}.".format(
                 object_name_list, relationship_class_name
             ),
-            id=current_items_by_obj_lst[class_id, join_object_id_list],
+            id=current_items_by_obj_lst[class_id, object_id_list],
         )
 
 
@@ -315,7 +314,7 @@ def check_entity_group(item, current_items, entities):
         )
 
 
-def check_parameter_definition(item, current_items, entity_class_ids, parameter_value_lists):
+def check_parameter_definition(item, current_items, entity_class_ids, parameter_value_lists, list_values):
     """Check whether the insertion of a parameter definition item
     results in the violation of an integrity constraint.
 
@@ -323,7 +322,7 @@ def check_parameter_definition(item, current_items, entity_class_ids, parameter_
         item (dict): A parameter definition item to be checked.
         current_items (dict): A dictionary mapping tuples (entity_class_id, name) to ids of parameter definitions
             already in the database.
-        entity_class_ids (dict): A set of entity class ids in the database.
+        entity_class_ids (Iterable): A set of entity class ids in the database.
         parameter_value_lists (dict): A dictionary of value-lists in the database keyed by id.
 
     Raises:
@@ -338,25 +337,19 @@ def check_parameter_definition(item, current_items, entity_class_ids, parameter_
     if entity_class_id not in entity_class_ids:
         raise SpineIntegrityError(
             f"Entity class id for parameter definition '{name}' not found in the entity class "
-            f"ids of the current database."
+            "ids of the current database."
         )
     if (entity_class_id, name) in current_items:
         raise SpineIntegrityError(
             "There's already a parameter called {0} in entity class with id {1}.".format(name, entity_class_id),
             id=current_items[entity_class_id, name],
         )
-    parameter_value_list_id = item.get("parameter_value_list_id")
-    if parameter_value_list_id is not None and parameter_value_list_id not in parameter_value_lists:
-        raise SpineIntegrityError("Invalid parameter value list.")
-    default_value = item.get("default_value")
-    default_type = item.get("default_type")
-    try:
-        _ = from_database(default_value, default_type)
-    except ParameterValueFormatError as err:
-        raise SpineIntegrityError("Invalid default value '{}': {}".format(default_value, err))
+    replace_default_values_with_list_references(item, parameter_value_lists, list_values)
 
 
-def check_parameter_value(item, current_items, parameter_definitions, entities, parameter_value_lists, alternatives):
+def check_parameter_value(
+    item, current_items, parameter_definitions, entities, parameter_value_lists, list_values, alternatives
+):
     """Check whether the insertion of a parameter value item results in the violation of an integrity constraint.
 
     Args:
@@ -366,6 +359,8 @@ def check_parameter_value(item, current_items, parameter_definitions, entities,
         parameter_definitions (dict): A dictionary of parameter definition items in the database keyed by id.
         entities (dict): A dictionary of entity items already in the database keyed by id.
         parameter_value_lists (dict): A dictionary of value-lists in the database keyed by id.
+        list_values (dict): A dictionary of list-values in the database keyed by id.
+        alternatives (set)
 
     Raises:
         SpineIntegrityError: if the insertion of the item violates an integrity constraint.
@@ -378,27 +373,9 @@ def check_parameter_value(item, current_items, parameter_definitions, entities,
         parameter_definition = parameter_definitions[parameter_definition_id]
     except KeyError:
         raise SpineIntegrityError("Parameter not found.")
-    value = item.get("value")
-    value_type = item.get("type")
     alt_id = item.get("alternative_id")
     if alt_id not in alternatives:
         raise SpineIntegrityError("Alternative not found.")
-    try:
-        _ = from_database(value, value_type)
-    except ParameterValueFormatError as err:
-        raise SpineIntegrityError("Invalid value '{}': {}".format(value, err))
-    if value is not None:
-        parameter_value_list_id = parameter_definition["parameter_value_list_id"]
-        value_list = parameter_value_lists.get(parameter_value_list_id)
-        if value_list is not None:
-            value_list = value_list.split(";")
-            if str(value, "UTF8") not in value_list:
-                valid_values = ", ".join(value_list)
-                raise SpineIntegrityError(
-                    "The value '{}' is not a valid value for parameter '{}' (valid values are: {})".format(
-                        value, parameter_definition["name"], valid_values
-                    )
-                )
     entity_id = item.get("entity_id")
     if not entity_id:
         raise SpineIntegrityError("Missing object or relationship identifier.")
@@ -417,38 +394,106 @@ def check_parameter_value(item, current_items, parameter_definitions, entities,
             "The value of parameter '{}' for entity '{}' is already specified.".format(parameter_name, entity_name),
             id=current_items[entity_id, parameter_definition_id, alt_id],
         )
+    replace_parameter_values_with_list_references(item, parameter_definitions, parameter_value_lists, list_values)
 
 
-def check_wide_parameter_value_list(wide_item, current_items):
-    """Check whether the insertion of a parameter value-list item
-    results in the violation of an integrity constraint.
+def replace_default_values_with_list_references(item, parameter_value_lists, list_values):
+    parameter_value_list_id = item.get("parameter_value_list_id")
+    return _replace_values_with_list_references(
+        "parameter_definition", item, parameter_value_list_id, parameter_value_lists, list_values
+    )
+
+
+def replace_parameter_values_with_list_references(item, parameter_definitions, parameter_value_lists, list_values):
+    parameter_definition_id = item["parameter_definition_id"]
+    parameter_definition = parameter_definitions[parameter_definition_id]
+    parameter_value_list_id = parameter_definition["parameter_value_list_id"]
+    return _replace_values_with_list_references(
+        "parameter_value", item, parameter_value_list_id, parameter_value_lists, list_values
+    )
+
+
+def _replace_values_with_list_references(item_type, item, parameter_value_list_id, parameter_value_lists, list_values):
+    if parameter_value_list_id is None:
+        return False
+    if parameter_value_list_id not in parameter_value_lists:
+        raise SpineIntegrityError("Parameter value list not found.")
+    value_id_list = parameter_value_lists[parameter_value_list_id]
+    if value_id_list is None:
+        raise SpineIntegrityError("Parameter value list is empty!")
+    value_key, type_key = {
+        "parameter_value": ("value", "type"),
+        "parameter_definition": ("default_value", "default_type"),
+    }[item_type]
+    value = item.get(value_key)
+    value_type = item.get(type_key)
+    try:
+        parsed_value = from_database(value, value_type)
+    except ParameterValueFormatError as err:
+        raise SpineIntegrityError(f"Invalid {value_key} '{value}': {err}") from None
+    if parsed_value is None:
+        return False
+    value_id_list = [int(id_) for id_ in value_id_list.split(",")]
+    list_value_id = next((id_ for id_ in value_id_list if list_values.get(id_) == parsed_value), None)
+    if list_value_id is None:
+        valid_values = ", ".join([f"'{list_values.get(id_)}'" for id_ in value_id_list])
+        raise SpineIntegrityError(
+            f"Invalid {value_key} '{parsed_value}' - it should be one from the parameter value list: {valid_values}."
+        )
+    item[value_key] = str(list_value_id).encode("UTF8")
+    item[type_key] = "list_value_ref"
+    return True
+
+
+def check_parameter_value_list(item, current_items):
+    """Check whether the insertion of a parameter value-list item results in the violation of an integrity constraint.
 
     Args:
-        wide_item (dict): A wide parameter value-list item to be checked.
+        item (dict): A parameter value-list item to be checked.
         current_items (dict): A dictionary mapping names to ids of parameter value-lists already in the database.
 
     Raises:
         SpineIntegrityError: if the insertion of the item violates an integrity constraint.
     """
     try:
-        name = wide_item["name"]
+        name = item["name"]
     except KeyError:
         raise SpineIntegrityError("Missing parameter value list name.")
     if name in current_items:
         raise SpineIntegrityError(
             "There can't be more than one parameter value_list called '{}'.".format(name), id=current_items[name]
         )
-    try:
-        value_list = wide_item["value_list"]
-    except KeyError:
-        raise SpineIntegrityError("Missing list of values.")
-    if len(value_list) != len(set(value_list)):
-        raise SpineIntegrityError("Values must be unique.")
-    for value in value_list:
-        try:
-            _ = from_database(value, value_type=None)
-        except ParameterValueFormatError as err:
-            raise SpineIntegrityError("Invalid value '{}': {}".format(value, err))
+
+
+def check_list_value(item, list_names_by_id, list_value_ids_by_index, list_value_ids_by_value):
+    """Check whether the insertion of a list value item results in the violation of an integrity constraint.
+
+    Args:
+        item (dict): A list value item to be checked.
+        list_names_by_id (dict): Mapping parameter value list ids to names.
+        list_value_ids_by_index (dict): Mapping tuples (list id, index) to ids of existing list values.
+        list_value_ids_by_value (dict): Mapping tuples (list id, type, value) to ids of existing list values.
+
+    Raises:
+        SpineIntegrityError: if the insertion of the item violates an integrity constraint.
+    """
+    keys = {"parameter_value_list_id", "index", "value", "type"}
+    missing_keys = keys - item.keys()
+    if missing_keys:
+        raise SpineIntegrityError(f"Missing keys: {', '.join(missing_keys)}.")
+    list_id = item["parameter_value_list_id"]
+    list_name = list_names_by_id.get(list_id)
+    if list_name is None:
+        raise SpineIntegrityError("Unknown parameter value list identifier.")
+    index = item["index"]
+    type_ = item["type"]
+    value = item["value"]
+    dup_id = list_value_ids_by_index.get((list_id, index))
+    if dup_id is not None:
+        raise SpineIntegrityError(f"'{list_name}' already has the index '{index}'.", id=dup_id)
+    dup_id = list_value_ids_by_value.get((list_id, type_, value))
+    if dup_id is not None:
+        raise SpineIntegrityError(f"'{list_name}' already has the value '{from_database(value, type_)}'.", id=dup_id)
 
 
 def check_tool(item, current_items):
@@ -540,3 +585,61 @@ def check_tool_feature_method(item, current_items, tool_features, parameter_valu
         raise SpineIntegrityError("Feature and parameter value list don't match.")
     if method_index not in parameter_value_list["value_index_list"]:
         raise SpineIntegrityError("Invalid method for tool feature.")
+
+
+def check_metadata(item, metadata):
+    """Check whether the entity metadata item violates an integrity constraint.
+
+    Args:
+        item (dict): An entity metadata item to be checked.
+        metadata (dict): Mapping from metadata name and value to metadata id.
+
+    Raises:
+        SpineIntegrityError: if the item violates an integrity constraint.
+    """
+    keys = {"name", "value"}
+    missing_keys = keys - item.keys()
+    if missing_keys:
+        raise SpineIntegrityError(f"Missing keys: {', '.join(missing_keys)}.")
+
+
+def check_entity_metadata(item, entities, metadata):
+    """Check whether the entity metadata item violates an integrity constraint.
+
+    Args:
+        item (dict): An entity metadata item to be checked.
+        entities (set of int): Available entity ids.
+        metadata (set of int): Available metadata ids.
+
+    Raises:
+        SpineIntegrityError: if the item violates an integrity constraint.
+    """
+    keys = {"entity_id", "metadata_id"}
+    missing_keys = keys - item.keys()
+    if missing_keys:
+        raise SpineIntegrityError(f"Missing keys: {', '.join(missing_keys)}.")
+    if item["entity_id"] not in entities:
+        raise SpineIntegrityError("Unknown entity identifier.")
+    if item["metadata_id"] not in metadata:
+        raise SpineIntegrityError("Unknown metadata identifier.")
+
+
+def check_parameter_value_metadata(item, values, metadata):
+    """Check whether the parameter value metadata item violates an integrity constraint.
+
+    Args:
+        item (dict): An entity metadata item to be checked.
+        values (set of int): Available parameter value ids.
+        metadata (set of int): Available metadata ids.
+
+    Raises:
+        SpineIntegrityError: if the item violates an integrity constraint.
+    """
+    keys = {"parameter_value_id", "metadata_id"}
+    missing_keys = keys - item.keys()
+    if missing_keys:
+        raise SpineIntegrityError(f"Missing keys: {', '.join(missing_keys)}.")
+    if item["parameter_value_id"] not in values:
+        raise SpineIntegrityError("Unknown parameter value identifier.")
+    if item["metadata_id"] not in metadata:
+        raise SpineIntegrityError("Unknown metadata identifier.")
