@@ -20,12 +20,11 @@ from datetime import datetime
 from sqlalchemy import func, Table, Column, Integer, String, null
 from sqlalchemy.exc import DBAPIError
 from .exception import SpineDBAPIError
-from .helpers import get_relationship_entity_class_items, get_relationship_entity_items, get_parameter_value_list_items
+from .helpers import get_relationship_entity_class_items, get_relationship_entity_items
 
 
 class DatabaseMappingAddMixin:
-    """Provides methods to perform ``INSERT`` operations over a Spine db.
-    """
+    """Provides methods to perform ``INSERT`` operations over a Spine db."""
 
     def __init__(self, *args, **kwargs):
         """Initialize class."""
@@ -43,6 +42,7 @@ class DatabaseMappingAddMixin:
                 Column("parameter_definition_id", Integer, server_default=null()),
                 Column("parameter_value_id", Integer, server_default=null()),
                 Column("parameter_value_list_id", Integer, server_default=null()),
+                Column("list_value_id", Integer, server_default=null()),
                 Column("alternative_id", Integer, server_default=null()),
                 Column("scenario_id", Integer, server_default=null()),
                 Column("scenario_alternative_id", Integer, server_default=null()),
@@ -61,7 +61,7 @@ class DatabaseMappingAddMixin:
                 next_id = Table("next_id", self._metadata, autoload=True)
         self._next_id = next_id
 
-    def _items_and_ids(self, tablename, *items):
+    def _add_commit_id_and_ids(self, tablename, *items):
         if not items:
             return [], set()
         fieldname = {
@@ -73,6 +73,7 @@ class DatabaseMappingAddMixin:
             "parameter_definition": "parameter_definition_id",
             "parameter_value": "parameter_value_id",
             "parameter_value_list": "parameter_value_list_id",
+            "list_value": "list_value_id",
             "alternative": "alternative_id",
             "scenario": "scenario_id",
             "scenario_alternative": "scenario_alternative_id",
@@ -100,13 +101,9 @@ class DatabaseMappingAddMixin:
             new_next_id = next_id + len(items)
             self._checked_execute(stmt, {"user": self.username, "date": datetime.utcnow(), fieldname: new_next_id})
         ids = list(range(next_id, new_next_id))
-        items_to_add = list()
-        append_item = items_to_add.append
         for id_, item in zip(ids, items):
             item["commit_id"] = self.make_commit_id()
             item["id"] = id_
-            append_item(item)
-        return items_to_add, set(ids)
 
     def _readd_items(self, tablename, *items):
         """Add known items to database."""
@@ -134,27 +131,31 @@ class DatabaseMappingAddMixin:
                 if the insertion of one of the items violates an integrity constraint.
             return_dups (bool): Whether or not already existing and duplicated entries should also be returned.
             return_items (bool): Return full items rather than just ids
-            cache (dict): A dict mapping table names to a list of dictionary items, to use as db replacement
+            cache (dict, optional): A dict mapping table names to a list of dictionary items, to use as db replacement
                 for queries
             readd (bool): Readds items directly
 
         Returns:
-            set: ids or items succesfully added
+            set: ids or items successfully added
             list(SpineIntegrityError): found violations
         """
         if readd:
             self._readd_items(tablename, *items)
             return
         if check:
-            checked_items, intgr_error_log = self.check_items_for_insert(tablename, *items, strict=strict, cache=cache)
+            checked_items, intgr_error_log = self.check_items(
+                tablename, *items, for_update=False, strict=strict, cache=cache
+            )
         else:
-            checked_items, intgr_error_log = items, []
-        result = self._add_items(tablename, *checked_items, return_items=return_items)
-        if return_dups and not return_items:
-            result.update(set(x.id for x in intgr_error_log if x.id))
-        return result, intgr_error_log
+            checked_items, intgr_error_log = list(items), []
+        ids = self._add_items(tablename, *checked_items)
+        if return_items:
+            return checked_items, intgr_error_log
+        if return_dups:
+            ids.update(set(x.id for x in intgr_error_log if x.id))
+        return ids, intgr_error_log
 
-    def _add_items(self, tablename, *items, return_items=False):
+    def _add_items(self, tablename, *items):
         """Add items to database without checking integrity.
 
         Args:
@@ -166,20 +167,22 @@ class DatabaseMappingAddMixin:
         Returns:
             ids (set): added instances' ids
         """
-        items_to_add, ids = self._items_and_ids(tablename, *items)
-        for _ in self._do_add_items(tablename, *items_to_add):
+        self._add_commit_id_and_ids(tablename, *items)
+        for _ in self._do_add_items(tablename, *items):
             pass
-        return items_to_add if return_items else ids
+        return {item["id"] for item in items}
 
     def _get_table_for_insert(self, tablename):
         """
-        Returns the Table object to perform the insertion. Subclasses can override this method to insert
-        to another table instead (e.g., diff...)
+        Returns the table name where to perform insertion.
+
+        Subclasses can override this method to insert to another table instead (e.g., diff...)
 
         Args:
-            tablename (str)
-        Returns:
-            Table
+            tablename (str): target database table name
+
+        Yields:
+            str: database table name
         """
         return self._metadata.tables[tablename]
 
@@ -200,11 +203,11 @@ class DatabaseMappingAddMixin:
         actually need to insert records to more than one table.
 
         Args:
-            tablename (str):
-            items_to_add (list)
+            tablename (str): target database table name
+            items_to_add (list): items to add
 
-        Returns:
-            Generator
+        Yields:
+            tuple: database table name, items to add
         """
         if tablename == "object_class":
             oc_items_to_add = list()
@@ -265,11 +268,6 @@ class DatabaseMappingAddMixin:
                     or item.get("entity_class_id")
                 )
             yield ("parameter_value", items_to_add)
-        elif tablename == "parameter_value_list":
-            items_to_add_ = list()
-            for item in items_to_add:
-                items_to_add_ += get_parameter_value_list_items(item)
-            yield ("parameter_value_list", items_to_add_)
         else:
             yield (tablename, items_to_add)
 
@@ -291,9 +289,12 @@ class DatabaseMappingAddMixin:
     def add_parameter_values(self, *items, **kwargs):
         return self.add_items("parameter_value", *items, **kwargs)
 
-    def add_wide_parameter_value_lists(self, *items, **kwargs):
+    def add_parameter_value_lists(self, *items, **kwargs):
         return self.add_items("parameter_value_list", *items, **kwargs)
 
+    def add_list_values(self, *items, **kwargs):
+        return self.add_items("list_value", *items, **kwargs)
+
     def add_features(self, *items, **kwargs):
         return self.add_items("feature", *items, **kwargs)
 
@@ -318,6 +319,74 @@ class DatabaseMappingAddMixin:
     def add_entity_groups(self, *items, **kwargs):
         return self.add_items("entity_group", *items, **kwargs)
 
+    def add_metadata(self, *items, **kwargs):
+        return self.add_items("metadata", *items, **kwargs)
+
+    def add_entity_metadata(self, *items, **kwargs):
+        return self.add_items("entity_metadata", *items, **kwargs)
+
+    def _get_or_add_metadata_ids_for_items(self, *items, check, strict, cache):
+        metadata_ids = {}
+        for entry in cache.get("metadata", {}).values():
+            metadata_ids.setdefault(entry.name, {})[entry.value] = entry.id
+        metadata_to_add = []
+        items_missing_metadata_ids = {}
+        for item in items:
+            existing_values = metadata_ids.get(item["metadata_name"])
+            existing_id = existing_values.get(item["metadata_value"]) if existing_values is not None else None
+            if existing_values is None or existing_id is None:
+                metadata_to_add.append({"name": item["metadata_name"], "value": item["metadata_value"]})
+                items_missing_metadata_ids.setdefault(item["metadata_name"], {})[item["metadata_value"]] = item
+            else:
+                item["metadata_id"] = existing_id
+        added_metadata, errors = self.add_items(
+            "metadata", *metadata_to_add, check=check, strict=strict, return_items=True, cache=cache
+        )
+        if errors:
+            return added_metadata, errors
+        new_metadata_ids = {}
+        for added in added_metadata:
+            new_metadata_ids.setdefault(added["name"], {})[added["value"]] = added["id"]
+        for metadata_name, value_to_item in items_missing_metadata_ids.items():
+            for metadata_value, item in value_to_item.items():
+                item["metadata_id"] = new_metadata_ids[metadata_name][metadata_value]
+        return added_metadata, errors
+
+    def _add_ext_item_metadata(self, table_name, *items, check=True, strict=False, return_items=False, cache=None):
+        # Note, that even though return_items can be False, it doesn't make much sense here because we'll be mixing
+        # metadata and entity metadata ids.
+        if cache is None:
+            cache = self.make_cache({table_name}, include_ancestors=True)
+        added_metadata, metadata_errors = self._get_or_add_metadata_ids_for_items(
+            *items, check=check, strict=strict, cache=cache
+        )
+        if metadata_errors:
+            if not return_items:
+                return added_metadata, metadata_errors
+            return {i["id"] for i in added_metadata}, metadata_errors
+        # We want to invalidate cache just in case because it might be missing metadata
+        # records that may have been added above.
+        added_item_metadata, item_errors = self.add_items(
+            table_name, *items, check=check, strict=strict, return_items=True, cache=None
+        )
+        errors = metadata_errors + item_errors
+        if not return_items:
+            return {i["id"] for i in added_metadata + added_item_metadata}, errors
+        return added_metadata + added_item_metadata, errors
+
+    def add_ext_entity_metadata(self, *items, check=True, strict=False, return_items=False, cache=None):
+        return self._add_ext_item_metadata(
+            "entity_metadata", *items, check=check, strict=strict, return_items=return_items, cache=cache
+        )
+
+    def add_parameter_value_metadata(self, *items, **kwargs):
+        return self.add_items("parameter_value_metadata", *items, **kwargs)
+
+    def add_ext_parameter_value_metadata(self, *items, check=True, strict=False, return_items=False, cache=None):
+        return self._add_ext_item_metadata(
+            "parameter_value_metadata", *items, check=check, strict=strict, return_items=return_items, cache=cache
+        )
+
     def _add_object_classes(self, *items):
         return self._add_items("object_class", *items)
 
@@ -336,9 +405,12 @@ class DatabaseMappingAddMixin:
     def _add_parameter_values(self, *items):
         return self._add_items("parameter_value", *items)
 
-    def _add_wide_parameter_value_lists(self, *items):
+    def _add_parameter_value_lists(self, *items):
         return self._add_items("parameter_value_list", *items)
 
+    def _add_list_values(self, *items):
+        return self._add_items("list_value", *items)
+
     def _add_features(self, *items):
         return self._add_items("feature", *items)
 
@@ -378,7 +450,7 @@ class DatabaseMappingAddMixin:
         :raises SpineIntegrityError: if the insertion of the item violates an integrity constraint.
 
         :returns:
-            - **new_item** -- The item succesfully staged for insertion.
+            - **new_item** -- The item successfully staged for insertion.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -392,7 +464,7 @@ class DatabaseMappingAddMixin:
         :raises SpineIntegrityError: if the insertion of the item violates an integrity constraint.
 
         :returns:
-            - **new_item** -- The item succesfully staged for insertion.
+            - **new_item** -- The item successfully staged for insertion.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -406,7 +478,7 @@ class DatabaseMappingAddMixin:
         :raises SpineIntegrityError: if the insertion of the item violates an integrity constraint.
 
         :returns:
-            - **new_item** -- The item succesfully staged for insertion.
+            - **new_item** -- The item successfully staged for insertion.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -420,7 +492,7 @@ class DatabaseMappingAddMixin:
         :raises SpineIntegrityError: if the insertion of the item violates an integrity constraint.
 
         :returns:
-            - **new_item** -- The item succesfully staged for insertion.
+            - **new_item** -- The item successfully staged for insertion.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -434,7 +506,7 @@ class DatabaseMappingAddMixin:
         :raises SpineIntegrityError: if the insertion of the item violates an integrity constraint.
 
         :returns:
-            - **new_item** -- The item succesfully staged for insertion.
+            - **new_item** -- The item successfully staged for insertion.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -448,7 +520,7 @@ class DatabaseMappingAddMixin:
         :raises SpineIntegrityError: if the insertion of the item violates an integrity constraint.
 
         :returns:
-            - **new_item** -- The item succesfully staged for insertion.
+            - **new_item** -- The item successfully staged for insertion.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -460,7 +532,7 @@ class DatabaseMappingAddMixin:
         """Stage an object class item for insertion if it doesn't already exists in the db.
 
         :returns:
-            - **item** -- The item succesfully staged for insertion or already existing.
+            - **item** -- The item successfully staged for insertion or already existing.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -472,7 +544,7 @@ class DatabaseMappingAddMixin:
         """Stage an object item for insertion if it doesn't already exists in the db.
 
         :returns:
-            - **item** -- The item succesfully staged for insertion or already existing.
+            - **item** -- The item successfully staged for insertion or already existing.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
@@ -484,7 +556,7 @@ class DatabaseMappingAddMixin:
         """Stage a parameter definition item for insertion if it doesn't already exists in the db.
 
         :returns:
-            - **item** -- The item succesfully staged for insertion or already existing.
+            - **item** -- The item successfully staged for insertion or already existing.
 
         :rtype: :class:`~sqlalchemy.util.KeyedTuple`
         """
