@@ -57,11 +57,14 @@ def get_mapped_data(
             raise TypeError(f"mapping must be a dict or ImportMapping subclass, instead got: {type(mapping).__name__}")
     if column_convert_fns is None:
         column_convert_fns = {}
+    default_column_convert_fn = column_convert_fns[max(column_convert_fns)] if column_convert_fns else lambda x: x
     if row_convert_fns is None:
         row_convert_fns = {}
     mapped_data = {}
     errors = []
     rows = list(data_source)
+    if not rows:
+        return mapped_data, errors
     for mapping in mappings:
         read_state = {}
         mapping = deepcopy(mapping)
@@ -90,6 +93,7 @@ def get_mapped_data(
         # If there are only pivoted mappings, we can just feed the unpivoted rows
         if not non_pivoted:
             # Reposition pivoted mappings:
+            last.position = -1
             for k, m in enumerate(pivoted):
                 m.position = k
             for k, row in enumerate(unpivoted_rows):
@@ -102,6 +106,7 @@ def get_mapped_data(
         # Reposition mappings:
         # - The last mapping (typically, parameter value) will read from the last position in the row
         # - The pivoted mappings will read from positions to the left of that
+        k = None
         last.position = -1
         for k, m in enumerate(reversed(pivoted)):
             m.position = -(k + 2)
@@ -110,13 +115,18 @@ def get_mapped_data(
         last_pivoted_row_pos = max(pivoted_pos, default=0) + 1
         last_non_pivoted_column_pos = max(non_pivoted_pos, default=0) + 1
         start_pos = max(mapping.read_start_row, last_pivoted_row_pos)
+        min_row_length = max(unpivoted_column_pos)
         for i, row in enumerate(rows[start_pos:]):
+            if len(row) < min_row_length + 1:
+                error = f"Could not process incomplete row {i + 1}"
+                errors.append(error)
+                continue
             if not _is_valid_row(row[:last_non_pivoted_column_pos]):
                 continue
-            row = _convert_row(row, column_convert_fns, start_pos + i, errors)
+            row = _convert_row(row, column_convert_fns, start_pos + i, errors, default_column_convert_fn)
             non_pivoted_row = row[:last_non_pivoted_column_pos]
             for column_pos, unpivoted_row in zip(unpivoted_column_pos, unpivoted_rows):
-                if not _is_valid_row(unpivoted_row) or row[column_pos] is None:
+                if not _is_valid_row(unpivoted_row):
                     continue
                 unpivoted_row = _convert_row(unpivoted_row, row_convert_fns, k, errors)
                 full_row = non_pivoted_row + unpivoted_row
@@ -130,13 +140,13 @@ def _is_valid_row(row):
     return row is not None and not all(i is None for i in row)
 
 
-def _convert_row(row, convert_fns, row_number, errors):
+def _convert_row(row, convert_fns, row_number, errors, default_convert_fn=lambda x: x):
     new_row = []
     for j, item in enumerate(row):
         if item is None:
             new_row.append(item)
             continue
-        convert_fn = convert_fns.get(j, lambda x: x)
+        convert_fn = convert_fns.get(j, default_convert_fn)
         try:
             item = convert_fn(item)
         except (ValueError, ParameterValueFormatError):
@@ -163,6 +173,9 @@ def _split_mapping(mapping):
     non_pivoted = []
     pivoted_from_header = []
     for m in flattened:
+        if pivoted and m is flattened[-1]:
+            # If any other mapping is pivoted, ignore last mapping's position
+            break
         if m.position == Position.header and m.value is None:
             pivoted_from_header.append(m)
             continue
@@ -176,7 +189,7 @@ def _split_mapping(mapping):
 
 
 def _unpivot_rows(rows, data_header, pivoted, non_pivoted, pivoted_from_header, skip_columns):
-    """Upivots rows.
+    """Unpivots rows.
 
     Args:
         rows (list(list)): Source table rows
@@ -195,7 +208,7 @@ def _unpivot_rows(rows, data_header, pivoted, non_pivoted, pivoted_from_header,
     pivoted_pos = [-(m.position + 1) for m in pivoted]  # (-1) -> (0), (-2) -> (1), (-3) -> (2), etc.
     non_pivoted_pos = [m.position for m in non_pivoted]
     # Collect pivoted rows
-    pivoted_rows = [rows[pos] for pos in pivoted_pos]
+    pivoted_rows = [rows[pos] for pos in pivoted_pos] if non_pivoted_pos else rows
     # Prepend as many headers as needed
     for m in pivoted_from_header:
         pivoted.insert(0, m)
@@ -208,7 +221,15 @@ def _unpivot_rows(rows, data_header, pivoted, non_pivoted, pivoted_from_header,
     pivoted_rows = [[item for k, item in enumerate(row) if k not in skip_pos] for row in pivoted_rows]
     # Unpivot
     unpivoted_rows = [list(row) for row in zip(*pivoted_rows)]
-    unpivoted_column_pos = [k for k in range(len(rows[0])) if k not in skip_pos]
+    if not non_pivoted_pos:
+        last_pivoted_position = max(pivoted_pos)
+        expanded_pivoted_rows = []
+        for row in unpivoted_rows:
+            head = row[: last_pivoted_position + 1]
+            for data in row[last_pivoted_position + 1 :]:
+                expanded_pivoted_rows.append(head + [data])
+        unpivoted_rows = expanded_pivoted_rows
+    unpivoted_column_pos = [k for k in range(len(rows[0])) if k not in skip_pos] if rows else []
     return unpivoted_rows, pivoted_pos, non_pivoted_pos, unpivoted_column_pos
 
 
@@ -220,22 +241,34 @@ def _make_parameter_values(mapped_data):
         "relationship_parameters": 2,
     }
     for key, value_pos in parameter_value_pos.items():
-        for row in mapped_data.get(key, []):
+        rows = mapped_data.get(key)
+        if rows is None:
+            continue
+        parameter_value_data = []
+        for row in rows:
             try:
                 value = row[value_pos]
             except IndexError:
+                parameter_value_data.append(row)
                 continue
             if isinstance(value, dict):
+                if "data" not in value:
+                    continue
                 row[value_pos] = _parameter_value_from_dict(value)
-            if isinstance(value, str):
+            elif isinstance(value, str):
                 try:
                     row[value_pos] = from_database(*split_value_and_type(value))
                 except ParameterValueFormatError:
                     pass
+            parameter_value_data.append(row)
+        mapped_data[key] = parameter_value_data
 
 
 def _parameter_value_from_dict(d):
-    index_names = d.get("index_names", [""])
+    mapped_index_names = d.get("index_names", {0: ""})
+    index_names = (max(mapped_index_names) + 1) * [""]
+    for i, name in mapped_index_names.items():
+        index_names[i] = name
     if d["type"] == "map":
         map_ = _table_to_map(d["data"], compress=d.get("compress", False))
         if index_names != [""]:
@@ -291,7 +324,11 @@ def _apply_index_names(map_, index_names):
         map_ (Map): target Map.
         index_names (Sequence of str): index names, one for each Map depth
     """
-    map_.index_name = index_names[0]
+    name = index_names[0]
+    if name:
+        map_.index_name = index_names[0]
+    if len(index_names) == 1:
+        return
     for v in map_.values:
         if isinstance(v, Map):
             _apply_index_names(v, index_names[1:])
