@@ -90,8 +90,19 @@ def check_validity(root_mapping):
 
 
 class ImportMapping(Mapping):
-    def __init__(self, position, value=None, skip_columns=None, read_start_row=0):
-        super().__init__(position, value=value)
+    """Base class for import mappings."""
+
+    def __init__(self, position, value=None, skip_columns=None, read_start_row=0, filter_re=""):
+        """
+        Args:
+            position (int or Position): what to map in the source table
+            value (Any, optional): fixed value
+            skip_columns (Iterable of int, optional): index of columns that should be skipped;
+                useful when source is pivoted
+            read_start_row (int): at which source row importing should start
+            filter_re (str): regular expression for filtering
+        """
+        super().__init__(position, value, filter_re)
         self._skip_columns = None
         self._read_start_row = None
         self.skip_columns = skip_columns
@@ -196,7 +207,24 @@ class ImportMapping(Mapping):
             return 0
         return self.parent.rank + 1
 
+    def _filter_accepts_row(self, source_row):
+        """Whether or not the row passes the filter for this mapping."""
+        if self.position == Position.hidden and self.value is None:
+            return True
+        if self._filter_re is None:
+            return True
+        source_data = self._data(source_row)
+        return self._filter_re.search(str(source_data)) is not None
+
+    def filter_accepts_row(self, source_row):
+        """Whether or not the row passes the filter for all mappings in the hierarchy."""
+        return self._filter_accepts_row(source_row) and (
+            self.child is None or self.child.filter_accepts_row(source_row)
+        )
+
     def import_row(self, source_row, state, mapped_data, errors=None):
+        if not self.filter_accepts_row(source_row):
+            return
         if errors is None:
             errors = []
         if not (self.position == Position.hidden and self.value is None):
@@ -252,27 +280,28 @@ class ImportMapping(Mapping):
         return d
 
     @classmethod
-    def reconstruct(cls, position, mapping_dict):
+    def reconstruct(cls, position, value, skip_columns, read_start_row, filter_re, mapping_dict):
         """
         Reconstructs mapping.
 
         Args:
             position (int or Position, optional): mapping's position
+            value (Any): fixed value
+            skip_columns (Iterable of Int, optional): skipped columns
+            read_start_row (int): first source row to read
+            filter_re (str): filter regular expression
             mapping_dict (dict): serialized mapping
 
         Returns:
             Mapping: reconstructed mapping
         """
-        value = mapping_dict.get("value")
-        skip_columns = mapping_dict.get("skip_columns")
-        read_start_row = mapping_dict.get("read_start_row", 0)
-        mapping = cls(position, value=value, skip_columns=skip_columns, read_start_row=read_start_row)
+        mapping = cls(position, value, skip_columns, read_start_row, filter_re)
         return mapping
 
 
 class ImportObjectsMixin:
-    def __init__(self, position, value=None, skip_columns=None, read_start_row=0, import_objects=False):
-        super().__init__(position, value, skip_columns, read_start_row)
+    def __init__(self, position, value=None, skip_columns=None, read_start_row=0, filter_re="", import_objects=False):
+        super().__init__(position, value, skip_columns, read_start_row, filter_re)
         self.import_objects = import_objects
 
     def to_dict(self):
@@ -282,24 +311,17 @@ class ImportObjectsMixin:
         return d
 
     @classmethod
-    def reconstruct(cls, position, mapping_dict):
-        value = mapping_dict.get("value")
-        skip_columns = mapping_dict.get("skip_columns")
-        read_start_row = mapping_dict.get("read_start_row", 0)
+    def reconstruct(cls, position, value, skip_columns, read_start_row, filter_re, mapping_dict):
         import_objects = mapping_dict.get("import_objects", False)
-        mapping = cls(
-            position,
-            value=value,
-            skip_columns=skip_columns,
-            read_start_row=read_start_row,
-            import_objects=import_objects,
-        )
+        mapping = cls(position, value, skip_columns, read_start_row, filter_re, import_objects)
         return mapping
 
 
 class IndexedValueMixin:
-    def __init__(self, position, value=None, skip_columns=None, read_start_row=0, compress=False, options=None):
-        super().__init__(position, value, skip_columns, read_start_row)
+    def __init__(
+        self, position, value=None, skip_columns=None, read_start_row=0, filter_re="", compress=False, options=None
+    ):
+        super().__init__(position, value, skip_columns, read_start_row, filter_re)
         if options is None:
             options = {}
         self.compress = compress
@@ -314,20 +336,10 @@ class IndexedValueMixin:
         return d
 
     @classmethod
-    def reconstruct(cls, position, mapping_dict):
-        value = mapping_dict.get("value")
-        skip_columns = mapping_dict.get("skip_columns")
-        read_start_row = mapping_dict.get("read_start_row", 0)
+    def reconstruct(cls, position, value, skip_columns, read_start_row, filter_re, mapping_dict):
         compress = mapping_dict.get("compress", False)
         options = mapping_dict.get("options")
-        mapping = cls(
-            position,
-            value=value,
-            skip_columns=skip_columns,
-            read_start_row=read_start_row,
-            compress=compress,
-            options=options,
-        )
+        mapping = cls(position, value, skip_columns, read_start_row, filter_re, compress, options)
         return mapping
 
 
@@ -505,7 +517,9 @@ class ParameterDefinitionMapping(ImportMapping):
             raise KeyError(ImportKey.CLASS_NAME)
         parameter_name = state[ImportKey.PARAMETER_NAME] = str(source_data)
         parameter_definition = state[ImportKey.PARAMETER_DEFINITION] = [class_name, parameter_name]
-        mapped_data.setdefault(map_key, list()).append(parameter_definition)
+        default_values = state.get(ImportKey.PARAMETER_DEFAULT_VALUES)
+        if default_values is None or tuple(parameter_definition) not in default_values:
+            mapped_data.setdefault(map_key, list()).append(parameter_definition)
 
 
 class ParameterDefaultValueMapping(ImportMapping):
@@ -551,18 +565,44 @@ class ParameterDefaultValueTypeMapping(IndexedValueMixin, ImportMapping):
             parameter_definition.append(value_list_name)
 
 
-class DefaultValueIndexNameMapping(ImportMapping):
+class IndexNameMappingBase(ImportMapping):
+    """Base class for index name mappings."""
+
+    _STATE_KEY = NotImplemented
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self._id = None
+
+    def _value_key(self, state):
+        raise NotImplementedError()
+
+    def _import_row(self, source_data, state, mapped_data):
+        values = state[self._STATE_KEY]
+        value = values[self._value_key(state)]
+        if self._id is None:
+            self._id = 0
+            current = self
+            while True:
+                if current.parent is None:
+                    break
+                current = current.parent
+                if isinstance(current, type(self)):
+                    self._id += 1
+        value.setdefault("index_names", {})[self._id] = source_data
+
+
+class DefaultValueIndexNameMapping(IndexNameMappingBase):
     """Maps default value index names.
 
     Cannot be used as the topmost mapping; must have a :class:`ParameterDefaultValueTypeMapping` as parent.
     """
 
     MAP_TYPE = "DefaultValueIndexName"
+    _STATE_KEY = ImportKey.PARAMETER_DEFAULT_VALUES
 
-    def _import_row(self, source_data, state, mapped_data):
-        values = state[ImportKey.PARAMETER_DEFAULT_VALUES]
-        value = values[_default_value_key(state)]
-        value.setdefault("index_names", []).append(source_data)
+    def _value_key(self, state):
+        return _default_value_key(state)
 
 
 class ParameterDefaultValueIndexMapping(ImportMapping):
@@ -710,18 +750,17 @@ class ParameterValueIndexMapping(ImportMapping):
         state.setdefault(ImportKey.PARAMETER_VALUE_INDEXES, []).append(index)
 
 
-class IndexNameMapping(ImportMapping):
+class IndexNameMapping(IndexNameMappingBase):
     """Maps index names for indexed parameter values.
 
     Cannot be used as the topmost mapping; must have an :class:`ParameterValueTypeMapping` as a parent.
     """
 
     MAP_TYPE = "IndexName"
+    _STATE_KEY = ImportKey.PARAMETER_VALUES
 
-    def _import_row(self, source_data, state, mapped_data):
-        values = state[ImportKey.PARAMETER_VALUES]
-        value = values[_parameter_value_key(state)]
-        value.setdefault("index_names", []).append(source_data)
+    def _value_key(self, state):
+        return _parameter_value_key(state)
 
 
 class ExpandedParameterValueMapping(ImportMapping):
@@ -832,8 +871,10 @@ class ScenarioAlternativeMapping(ImportMapping):
     MAP_TYPE = "ScenarioAlternative"
 
     def _import_row(self, source_data, state, mapped_data):
-        scenario = state[ImportKey.SCENARIO_NAME]
         alternative = str(source_data)
+        if not alternative:
+            return
+        scenario = state[ImportKey.SCENARIO_NAME]
         scen_alt = state[ImportKey.SCENARIO_ALTERNATIVE] = [scenario, alternative]
         mapped_data.setdefault("scenario_alternatives", list()).append(scen_alt)
 
@@ -1040,9 +1081,17 @@ def from_dict(serialized):
     flattened = list()
     for mapping_dict in serialized:
         position = mapping_dict["position"]
+        value = mapping_dict.get("value")
+        skip_columns = mapping_dict.get("skip_columns")
+        read_start_row = mapping_dict.get("read_start_row", 0)
+        filter_re = mapping_dict.get("filter_re", "")
         if isinstance(position, str):
             position = Position(position)
-        flattened.append(mappings[mapping_dict["map_type"]].reconstruct(position, mapping_dict))
+        flattened.append(
+            mappings[mapping_dict["map_type"]].reconstruct(
+                position, value, skip_columns, read_start_row, filter_re, mapping_dict
+            )
+        )
     return unflatten(flattened)
 
 
