@@ -52,13 +52,6 @@ _TIME_SERIES_DEFAULT_RESOLUTION = "1h"
 _TIME_SERIES_PLAIN_INDEX_UNIT = "m"
 
 
-class ParameterValueEncoder(json.JSONEncoder):
-    """A class to serialize Spine parameter values."""
-
-    def default(self, o):
-        return dict(type=o.type_(), **o.to_dict())
-
-
 def duration_to_relativedelta(duration):
     """
     Converts a duration to a relativedelta object.
@@ -128,25 +121,62 @@ def relativedelta_to_duration(delta):
     return "0h"
 
 
+def load_db_value(db_value, value_type=None):
+    """
+    Loads a database parameter value into a Python object using JSON.
+    Adds the "type" property to dicts representing complex types.
+
+    Args:
+        db_value (bytes, optional): a value in the database
+        value_type (str, optional): the type in case of complex ones
+
+    Returns:
+        Any: the parsed parameter value
+    """
+    if db_value is None:
+        return None
+    try:
+        parsed = json.loads(db_value)
+    except JSONDecodeError as err:
+        raise ParameterValueFormatError(f"Could not decode the value: {err}") from err
+    if isinstance(parsed, dict):
+        return {"type": value_type, **parsed}
+    return parsed
+
+
+def dump_db_value(parsed_value):
+    """
+    Dumps a Python object into a database parameter value using JSON.
+    Extracts the "type" property from dicts representing complex types.
+
+    Args:
+        parsed_value (Any): the Python object
+
+    Returns:
+        str: the database parameter value
+        str: the type
+    """
+    value_type = parsed_value.pop("type") if isinstance(parsed_value, dict) else None
+    db_value = json.dumps(parsed_value).encode("UTF8")
+    if isinstance(parsed_value, dict) and value_type is not None:
+        parsed_value["type"] = value_type
+    return db_value, value_type
+
+
 def from_database(database_value, value_type=None):
     """
-    Converts a parameter value from its database representation to a Python object.
+    Converts a parameter value from its database representation into an encoded Python object.
 
     Args:
         database_value (bytes, optional): a value in the database
         value_type (str, optional): the type in case of complex ones
 
     Returns:
-        the encoded parameter value
+        Any: the encoded parameter value
     """
-    if database_value is None:
-        return None
-    try:
-        parsed = json.loads(database_value)
-    except JSONDecodeError as err:
-        raise ParameterValueFormatError(f"Could not decode the value: {err}")
+    parsed = load_db_value(database_value, value_type)
     if isinstance(parsed, dict):
-        return from_dict(parsed, value_type=value_type)
+        return from_dict(parsed)
     if isinstance(parsed, bool):
         return parsed
     if isinstance(parsed, Number):
@@ -156,9 +186,9 @@ def from_database(database_value, value_type=None):
 
 def from_database_to_single_value(database_value, value_type):
     """
-    Converts a value from its database representation to single value.
+    Converts a value from its database representation into a single value.
 
-    Indexed values get converted to strings representing the value's type.
+    Indexed values get converted to their type string.
 
     Args:
         database_value (bytes): a value in the database
@@ -192,35 +222,34 @@ def from_database_to_dimension_count(database_value, value_type):
     return 0
 
 
-def to_database(value):
+def to_database(parsed_value):
     """
-    Converts a value object into its database representation.
+    Converts an encoded Python object into its database representation.
 
     Args:
-        value: a value to convert (can be a dict with a "type" key)
+        value: the value to convert. It can be the result of either ``load_db_value`` or ``from_database```.
 
     Returns:
         bytes: value's database representation as bytes
         str: the value type
     """
-    if hasattr(value, "to_database"):
-        return value.to_database()
-    value_type = value.get("type") if isinstance(value, dict) else None
-    db_value = json.dumps(value).encode("UTF8")
-    return db_value, value_type
+    if hasattr(parsed_value, "to_database"):
+        return parsed_value.to_database()
+    db_value = json.dumps(parsed_value).encode("UTF8")
+    return db_value, None
 
 
-def from_dict(value_dict, value_type=None):
+def from_dict(value_dict):
     """
-    Converts complex a (relationship) parameter value from its dictionary representation to a Python object.
+    Converts a complex (relationship) parameter value from its dictionary representation to a Python object.
 
     Args:
-        value_dict (dict): value's dictionary; a parsed JSON object
-        value_type (str, optional): value's type
+        value_dict (dict): value's dictionary; a parsed JSON object with the "type" key
 
     Returns:
         the encoded (relationship) parameter value
     """
+    value_type = value_dict["type"]
     try:
         if value_type == "date_time":
             return _datetime_from_database(value_dict["data"])
@@ -239,6 +268,51 @@ def from_dict(value_dict, value_type=None):
         raise ParameterValueFormatError(f'"{error.args[0]}" is missing in the parameter value description')
 
 
+def fix_conflict(new, old, on_conflict="merge"):
+    """Resolves conflicts between parameter values:
+
+    Args:
+        new (any): new parameter value to write
+        old (any): existing parameter value in the db
+        on_conflict (str): conflict resolution strategy:
+            - 'merge': Merge indexes if possible, otherwise replace
+            - 'replace': Replace old with new
+            - 'keep': keep old
+
+    Returns:
+        any: a parameter value with conflicts resolved
+    """
+    funcs = {"keep": lambda new, old: old, "replace": lambda new, old: new, "merge": merge}
+    func = funcs.get(on_conflict)
+    if func is None:
+        raise RuntimeError(
+            f"Invalid conflict resolution strategy {on_conflict}, valid strategies are {', '.join(funcs)}"
+        )
+    return func(new, old)
+
+
+def merge(value, other):
+    """Merges other into value, returns the result.
+    Args:
+        value (tuple): recipient value and type
+        other (tuple): other value and type
+
+    Returns:
+        tuple: value and type of merged value
+    """
+    parsed_value = from_database(*value)
+    if not hasattr(parsed_value, "merge"):
+        return value
+    parsed_other = from_database(*other)
+    return to_database(parsed_value.merge(parsed_other))
+
+
+def merge_parsed(parsed_value, parsed_other):
+    if not hasattr(parsed_value, "merge"):
+        return parsed_value
+    return parsed_value.merge(parsed_other)
+
+
 def _break_dictionary(data):
     """Converts {"index": value} style dictionary into (list(indexes), numpy.ndarray(values)) tuple."""
     if not isinstance(data, dict):
@@ -515,7 +589,7 @@ def _map_values_from_database(values_in_db):
         return list()
     values = list()
     for value_in_db in values_in_db:
-        value = from_dict(value_in_db, value_in_db["type"]) if isinstance(value_in_db, dict) else value_in_db
+        value = from_dict(value_in_db) if isinstance(value_in_db, dict) else value_in_db
         if value is not None and not isinstance(value, (float, bool, Duration, IndexedValue, str, DateTime)):
             raise ParameterValueFormatError(f'Unsupported value type for Map: "{type(value).__name__}".')
         values.append(value)
@@ -575,6 +649,11 @@ class DateTime:
             return NotImplemented
         return self._value == other._value
 
+    def __lt__(self, other):
+        if not isinstance(other, DateTime):
+            return NotImplemented
+        return self._value < other._value
+
     def __hash__(self):
         return hash(self._value)
 
@@ -710,6 +789,7 @@ class IndexedValue:
             index_name (str): index name
         """
         self._indexes = None
+        self._values = None
         self.index_name = index_name
 
     def __bool__(self):
@@ -746,7 +826,12 @@ class IndexedValue:
     @property
     def values(self):
         """Returns the data values."""
-        raise NotImplementedError()
+        return self._values
+
+    @values.setter
+    def values(self, values):
+        """Sets the values."""
+        self._values = values
 
     def get_value(self, index):
         """Returns the value at the given index."""
@@ -769,6 +854,17 @@ class IndexedValue:
         """
         raise NotImplementedError()
 
+    def merge(self, other):
+        if not isinstance(other, type(self)):
+            return self
+        new_indexes = np.unique(np.concatenate((self.indexes, other.indexes)))
+        new_indexes.sort(kind='mergesort')
+        _merge = lambda value, other: other if value is None else merge_parsed(value, other)
+        new_values = [_merge(self.get_value(index), other.get_value(index)) for index in new_indexes]
+        self.indexes = new_indexes
+        self.values = new_values
+        return self
+
 
 class Array(IndexedValue):
     """A one dimensional array with zero based indexing."""
@@ -790,13 +886,13 @@ class Array(IndexedValue):
         if any(not isinstance(x, value_type) for x in values):
             raise ParameterValueFormatError("Not all array's values are of the same type.")
         self.indexes = range(len(values))
-        self._values = list(values)
+        self.values = list(values)
         self._value_type = value_type
 
     def __eq__(self, other):
         if not isinstance(other, Array):
             return NotImplemented
-        return self._values == other._values and self.index_name == other.index_name
+        return np.array_equal(self._values, other._values) and self.index_name == other.index_name
 
     @staticmethod
     def type_():
@@ -826,11 +922,6 @@ class Array(IndexedValue):
         """Returns the type of array's elements."""
         return self._value_type
 
-    @property
-    def values(self):
-        """See base class."""
-        return self._values
-
 
 class IndexedNumberArray(IndexedValue):
     """
@@ -846,6 +937,11 @@ class IndexedNumberArray(IndexedValue):
             values (Sequence): array's values; index handling should be implemented by subclasses
         """
         super().__init__(index_name)
+        self.values = values
+
+    @IndexedValue.values.setter
+    def values(self, values):
+        """Sets the values."""
         if not isinstance(values, np.ndarray) or not values.dtype == np.dtype(float):
             values = np.array(values, dtype=float)
         self._values = values
@@ -858,11 +954,6 @@ class IndexedNumberArray(IndexedValue):
         """Return the database representation of the value."""
         raise NotImplementedError()
 
-    @property
-    def values(self):
-        """Returns the data values as numpy.ndarray."""
-        return self._values
-
 
 class TimeSeries(IndexedNumberArray):
     """An abstract base class for time series."""
@@ -915,6 +1006,68 @@ class TimeSeries(IndexedNumberArray):
         raise NotImplementedError()
 
 
+def _check_time_pattern_index(union_str):
+    """
+    Checks if a time pattern index has the right format.
+
+    Args:
+        union_str (str): The time pattern index to check. Generally assumed to be a union of interval intersections.
+
+    Raises:
+        ParameterValueFormatError: If the given string doesn't comply with time pattern spec.
+    """
+    if not union_str:
+        # We accept empty strings so we can add empty rows in the parameter value editor UI
+        return
+    union_dlm = ","
+    intersection_dlm = ";"
+    range_dlm = "-"
+    regexp = r"(Y|M|D|WD|h|m|s)"
+    for intersection_str in union_str.split(union_dlm):
+        for interval_str in intersection_str.split(intersection_dlm):
+            m = re.match(regexp, interval_str)
+            if m is None:
+                raise ParameterValueFormatError(
+                    f"Invalid interval {interval_str}, it should start with either Y, M, D, WD, h, m, or s."
+                )
+            key = m.group(0)
+            lower_upper_str = interval_str[len(key) :]
+            lower_upper = lower_upper_str.split(range_dlm)
+            if len(lower_upper) != 2:
+                raise ParameterValueFormatError(
+                    f"Invalid interval bounds {lower_upper_str}, it should be two integers separated by dash (-)."
+                )
+            lower_str, upper_str = lower_upper
+            try:
+                lower = int(lower_str)
+            except:
+                raise ParameterValueFormatError(f"Invalid lower bound {lower_str}, must be an integer.")
+            try:
+                upper = int(upper_str)
+            except:
+                raise ParameterValueFormatError(f"Invalid upper bound {upper_str}, must be an integer.")
+            if lower > upper:
+                raise ParameterValueFormatError(f"Lower bound {lower} can't be higher than upper bound {upper}.")
+
+
+class _TimePatternIndexes(_Indexes):
+    """An array of *checked* time pattern indexes."""
+
+    def __array_finalize__(self, obj):
+        """Checks indexes when building the array."""
+        for x in obj:
+            _check_time_pattern_index(x)
+        super().__array_finalize__(obj)
+
+    def __eq__(self, other):
+        return list(self) == list(other)
+
+    def __setitem__(self, position, index):
+        """Checks indexes when setting and item."""
+        _check_time_pattern_index(index)
+        super().__setitem__(position, index)
+
+
 class TimePattern(IndexedNumberArray):
     """Represents a time pattern (relationship) parameter value."""
 
@@ -948,7 +1101,7 @@ class TimePattern(IndexedNumberArray):
     @IndexedNumberArray.indexes.setter
     def indexes(self, indexes):
         """Sets the indexes."""
-        self._indexes = _Indexes(indexes, dtype=np.object_)
+        self._indexes = _TimePatternIndexes(indexes, dtype=np.object_)
 
     @staticmethod
     def type_():
@@ -995,7 +1148,7 @@ class TimeSeriesFixedResolution(TimeSeries):
         return (
             self._start == other._start
             and self._resolution == other._resolution
-            and np.all(self._values == other._values)
+            and np.array_equal(self._values, other._values, equal_nan=True)
             and self._ignore_year == other._ignore_year
             and self._repeat == other._repeat
             and self.index_name == other.index_name
@@ -1132,8 +1285,8 @@ class TimeSeriesVariableResolution(TimeSeries):
         if not isinstance(other, TimeSeriesVariableResolution):
             return NotImplemented
         return (
-            self._indexes == other._indexes
-            and np.all(self._values == other._values)
+            np.array_equal(self._indexes, other._indexes)
+            and np.array_equal(self._values, other._values, equal_nan=True)
             and self._ignore_year == other._ignore_year
             and self._repeat == other._repeat
             and self.index_name == other.index_name
@@ -1183,11 +1336,6 @@ class Map(IndexedValue):
             return NotImplemented
         return other._indexes == self._indexes and other._values == self._values and self.index_name == other.index_name
 
-    @property
-    def values(self):
-        """Map's values."""
-        return self._values
-
     def is_nested(self):
         """Returns True if any of the values is also a map."""
         return any(isinstance(value, Map) for value in self._values)
@@ -1369,30 +1517,29 @@ def _try_convert_to_container(map_):
 SUPPORTED_TYPES = (Duration, DateTime, float, str)
 
 
-_VALUE_TYPE_SEP = "::"
-"""Separator for combining parameter value and type into a string.
-This is mainly to keep using a single 'field' for the value in Qt models.
-For copy-paste, we also need it to be a string (otherwise a tuple would have worked).
-"""
-
-
-def join_value_and_type(value, value_type):
-    """Returns a string that combines given parameter value and value type.
+def join_value_and_type(db_value, db_type):
+    """Joins database value and type into a string.
+    The resulting string is a JSON string.
+    In case of complex types (duration, date_time, time_series, time_pattern, array, map),
+    the type is just added as top-level key.
 
     Args:
-        value (bytes)
-        value_type (str or NoneType)
+        db_value (bytes): database value
+        db_type (str, optional): value type
 
     Returns:
-        str
+        str: parameter value as JSON with an additional `type` field.
     """
-    if value_type is None:
-        value_type = ""
-    return str(value, "UTF8") + _VALUE_TYPE_SEP + value_type
+    try:
+        parsed = load_db_value(db_value, db_type)
+    except ParameterValueFormatError:
+        parsed = None
+    return json.dumps(parsed)
 
 
 def split_value_and_type(value_and_type):
-    """Splits the given string into parameter value and value type.
+    """Splits the given string into value and type.
+    The string must be the result of calling ``join_value_and_type`` or have the same form.
 
     Args:
         value_and_type (str)
@@ -1401,9 +1548,8 @@ def split_value_and_type(value_and_type):
         bytes
         str or NoneType
     """
-    if value_and_type is None:
-        return None, None
-    value, _, value_type = value_and_type.partition(_VALUE_TYPE_SEP)
-    if not value_type:
-        value_type = None
-    return bytes(value, "UTF8"), value_type
+    try:
+        parsed = json.loads(value_and_type)
+    except (TypeError, json.JSONDecodeError):
+        parsed = value_and_type
+    return dump_db_value(parsed)
